#!/usr/bin/python3
#
# Copyright (C) 2019 Google Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np, pandas as pd, scipy.linalg as linalg, plspm.util as util
from enum import Enum
from typing import Tuple


class _ModeA:

    def outer_weights_metric(self, data: pd.DataFrame, Z: pd.DataFrame, lv: str, mvs: list) -> pd.DataFrame:
        return (1 / data.shape[0]) * Z.loc[:, [lv]].T.dot(data.loc[:, mvs]).T

    def outer_weights_nonmetric(self, mv_grouped_by_lv: list, mv_grouped_by_lv_missing: list, Z: np.ndarray, lv: str,
                                correction: float) -> Tuple[np.ndarray, np.ndarray]:
        if lv in mv_grouped_by_lv_missing:
            weights = np.nansum(mv_grouped_by_lv[lv] * Z[:, np.newaxis], axis=0)
            weights = weights / np.sum(np.power(mv_grouped_by_lv_missing[lv] * Z[:, np.newaxis], 2), axis=0)
            Y = np.nansum(np.transpose(mv_grouped_by_lv[lv]) * weights[:, np.newaxis], axis=0)
            Y = Y / np.sum(np.power(np.transpose(mv_grouped_by_lv_missing[lv]) * weights[:, np.newaxis], 2), axis=0)
        else:   
            weights = np.dot(np.transpose(mv_grouped_by_lv[lv]), Z) / np.power(Z, 2).sum()
            Y = np.dot(mv_grouped_by_lv[lv], weights)
        Y = util.treat_numpy(Y) * correction
        return weights, Y


class _ModeB:

    def outer_weights_metric(self, data: pd.DataFrame, Z: pd.DataFrame, lv: str, mvs: list) -> pd.DataFrame:
        w, _, _, _ = linalg.lstsq(data.loc[:, mvs], Z.loc[:, [lv]])
        return pd.DataFrame(w, columns=[lv], index=mvs)

    def outer_weights_nonmetric(self, mv_grouped_by_lv: list, mv_grouped_by_lv_missing: list, Z: pd.DataFrame, lv: str,
                                correction: float) -> Tuple[np.ndarray, np.ndarray]:
        if lv in mv_grouped_by_lv_missing:
            raise Exception("Missing nonmetric data is not supported in mode B. LV with missing data: " + lv)
        weights, _, _, _ = linalg.lstsq(mv_grouped_by_lv[lv], Z)
        Y = np.dot(mv_grouped_by_lv[lv], weights)
        Y = util.treat_numpy(Y) * correction
        return weights, Y


class Mode(Enum):
    """
    Specify whether a given latent variable is reflective (mode A) or formative (mode B) with respect to its manifest variables.
    """
    A = _ModeA()
    B = _ModeB()



#!/usr/bin/python3
#
# Copyright (C) 2019 Google Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Tuple

import numpy as np, pandas as pd, plspm.config as c, statsmodels.api as sm, plspm.util as util, math
from plspm.scheme import Scheme
from plspm.mode import Mode

pd.options.mode.chained_assignment = None  # default='warn'


class _MetricWeights:
    """Internal class that calculates weights and scores when using metric data."""
    def __init__(self, data: pd.DataFrame, config: c.Config, correction: float):
        odm = config.odm()
        weight_factors = correction / data.dot(odm).std(axis=0)
        self.__mvs = list(odm.index)
        wf_diag = pd.DataFrame(np.diag(weight_factors), index=weight_factors.index, columns=weight_factors.index)
        weights = odm.dot(wf_diag)
        self.__w_old = weights.sum(axis=1).to_frame(name="weight")
        self.__data = data
        self.__config = config
        self.__weights = weights
        self.__correction = correction

    def iterate(self, inner_weight_calculator: Scheme) -> float:
        lvs = self.__config.lvs()
        y = self.__data.dot(self.__weights).reindex(lvs, axis=1)
        y = util.treat(y) / self.__correction
        inner_weights = pd.DataFrame(inner_weight_calculator.value.calculate(self.__config.path(), y.values), index=lvs, columns=lvs)
        Z = y.dot(inner_weights)
        for lv in list(lvs):
            mvs = self.__config.mvs(lv)
            weights = self.__config.mode(lv).value.outer_weights_metric(self.__data, Z, lv, mvs)
            self.__weights.loc[mvs, [lv]] = weights
        w_new = self.__weights.sum(axis=1).to_frame(name="weight")
        convergence = np.power(self.__w_old.abs() - w_new.abs(), 2).sum(axis=1).sum(axis=0)
        self.__w_old = w_new
        return convergence

    def calculate(self) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        weight_factors = 1 / (self.__data.dot(self.__weights).std(axis=0) / self.__correction)
        wf_diag = pd.DataFrame(np.diag(weight_factors), index=weight_factors.index, columns=weight_factors.index)
        weights = self.__weights.dot(wf_diag)
        y = self.__data.dot(weights)
        cor = pd.concat([self.__data, y], axis=1).corr().loc[list(self.__data), list(y)]
        odm = weights.apply(lambda x: x!= 0).astype(int)
        sign = lambda x : math.copysign(1.0, x)
        w_sign = (cor * odm).applymap(sign).sum(axis=0).apply(sign)
        if -1 in w_sign.values:
            w_sign = w_sign.apply(lambda x : -1 if x == 0 else x)
            w_sign_matrix = pd.DataFrame(np.diag(w_sign), index=w_sign.index, columns=w_sign.index)
            y = y.dot(w_sign_matrix)
        weights = pd.DataFrame(weights.sum(axis=1), index=self.__mvs, columns=["weight"])
        return self.__data, y, weights


class _NonmetricWeights:
    """Internal class that calculates weights and scores when using nonmetric data."""
    def __init__(self, data: pd.DataFrame, config: c.Config, correction: float):
        self.__mv_grouped_by_lv_initial = {}
        self.__mvs = []
        mv_grouped_by_lv = {}
        self.__mv_grouped_by_lv_missing = {}
        y = np.zeros((len(data.index), len(config.lvs())), dtype=np.float64)
        for i, lv in enumerate(config.lvs()):
            mvs = config.mvs(lv)
            self.__mvs.extend(mvs)
            mv_grouped_by_lv[lv] = data.filter(config.mvs(lv)).values.astype(np.float64)
            self.__mv_grouped_by_lv_initial[lv] = mv_grouped_by_lv[lv].copy()
            sizes = mv_grouped_by_lv[lv].shape[1]
            weight = [1 / np.sqrt(sizes)] * sizes
            if np.isnan(np.sum(mv_grouped_by_lv[lv])):
                self.__mv_grouped_by_lv_missing[lv] = 1 - np.isnan(mv_grouped_by_lv[lv])
                for j in range(len(data.index)):
                    numerator = np.nansum(mv_grouped_by_lv[lv][j, :] * weight)
                    denominator = np.power(weight * self.__mv_grouped_by_lv_missing[lv][j, :], 2).sum()
                    if denominator == 0:
                        raise ValueError("All mvs for lv " + lv + " in row " + str(j) + " are NaN.")
                    y[j, i] = numerator / denominator
            else:
                y[:, i] = np.dot(mv_grouped_by_lv[lv], weight)
        self.__weights = {}
        self.__y = y
        self.__mv_grouped_by_lv = mv_grouped_by_lv
        self.__config = config
        self.__correction = correction
        self.__index = data.index

    def iterate(self, inner_weight_calculator: Scheme) -> float:
        self.__betas = {}
        y_old = self.__y.copy()
        inner_weights = inner_weight_calculator.value.calculate(self.__config.path(), self.__y)
        Z = np.dot(self.__y, inner_weights)
        for i, lv in enumerate(list(self.__config.lvs())):
            for j, mv in enumerate(list(self.__config.mvs(lv))):
                self.__mv_grouped_by_lv[lv][:, j] = \
                    self.__config.scale(mv).value.scale(lv, mv, Z[:, i], self)
            self.__weights[lv], self.__y[:, i] = \
                self.__config.mode(lv).value.outer_weights_nonmetric(self.__mv_grouped_by_lv,
                                                                     self.__mv_grouped_by_lv_missing, Z[:, i], lv,
                                                                     self.__correction)
        return np.power(np.abs(y_old) - np.abs(self.__y), 2).sum()

    def calculate(self) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        lvs = self.__config.lvs()
        weights = pd.DataFrame(0, index=self.__mvs, columns=lvs)
        data_new = pd.DataFrame(0, index=self.__index, columns=self.__mvs)
        for lv in self.__config.lvs():
            mvs = self.__config.mvs(lv)
            weights.loc[mvs, [lv]] = self.__weights[lv]
            data_new.loc[:, mvs] = self.__mv_grouped_by_lv[lv]
        weight_factors = 1 / (data_new.dot(weights).std(axis=0, skipna=True) / self.__correction)
        wf_diag = pd.DataFrame(np.diag(weight_factors), index=lvs, columns=lvs)
        weights = weights.dot(wf_diag).sum(axis=1).to_frame(name="weight")
        return data_new, pd.DataFrame(self.__y, index=self.__index, columns=lvs), weights

    def get_Z_for_mode_b(self, lv, mv, z_by_lv):
        mv_index = self.__config.mv_index(lv, mv)
        if self.__config.mode(lv) != Mode.B or len(self.__config.mvs(lv)) == 1:
            return z_by_lv
        if lv not in self.__betas:
            exogenous = sm.add_constant(self.__mv_grouped_by_lv[lv])
            regression = sm.OLS(z_by_lv, exogenous).fit()
            self.__betas[lv] = regression.params[1:]
        correction = 1 / self.__betas[lv][mv_index]
        return correction * (z_by_lv - np.delete(self.__mv_grouped_by_lv[lv], mv_index, axis=1).dot(
            np.delete(self.__betas[lv], mv_index)))

    def correction(self) -> float:
        return self.__correction

    def mv_grouped_by_lv(self, lv: str, mv: str):
        return self.__mv_grouped_by_lv_initial[lv][:, self.__config.mv_index(lv, mv)]

    def dummies(self, mv: str) -> pd.DataFrame:
        return self.__config.dummies(mv)


class WeightsCalculatorFactory:
    """Internal class that is used to calculate weights and scores from the data using the model."""
    def __init__(self, config: c.Config, iterations: int, tolerance: float, correction: float, scheme: Scheme):
        self.__iterations = iterations
        self.__tolerance = tolerance
        self.__config = config
        self.__correction = correction
        self.__scheme = scheme

    def calculate(self, data: pd.DataFrame):
        """Internal method that performs the calculation to estimate weights and scores."""
        if self.__config.metric():
            calculator = _MetricWeights(data, self.__config, self.__correction)
        else:
            calculator = _NonmetricWeights(data, self.__config, self.__correction)

        iteration = 0
        while True:
            iteration += 1
            convergence = calculator.iterate(self.__scheme)
            if (convergence < self.__tolerance) or (iteration > self.__iterations):
                break
        if iteration > self.__iterations:
            raise Exception("Could not converge after " + str(iteration) + " iterations")
        return calculator.calculate()



#!/usr/bin/python3
#
# Copyright (C) 2019 Google Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import plspm.util as util, pandas as pd, numpy as np
from enum import Enum


class _Numeric:

    def scale(self, lv: str, mv: str, z_by_lv: pd.Series, weights) -> pd.DataFrame:
        data = weights.mv_grouped_by_lv(lv, mv)
        finite_elements = np.isfinite(data).sum()
        return util.treat_numpy(data) * np.sqrt(finite_elements / (finite_elements - 1))


class _Raw:

    def scale(self, lv: str, mv: str, z_by_lv: pd.Series, weights) -> pd.DataFrame:
        return weights.mv_grouped_by_lv(lv, mv)


class _Ordinal:

    def _quantify(self, dummies, z_by_lv):
        scaling = [0] * (len(dummies[0]))
        for n in range(len(dummies[0])):
            scaling[n] = np.sum(dummies[:, n] * z_by_lv)
            scaling[n] = scaling[n] / np.sum(dummies[:, n])
        return scaling

    def _ordinalize(self, scaling, dummies, z_by_lv, sign: int):
        while True:
            ncols = len(dummies[0])
            for n in range(len(dummies[0]) - 1):
                if np.sign(scaling[n] - scaling[n + 1]) == sign:
                    dummies[:, n + 1] = dummies[:, n] + dummies[:, n + 1]
                    dummies = np.delete(dummies, n, axis=1)
                    scaling = self._quantify(dummies, z_by_lv)
                    break
            if len(dummies[0]) == 1 or len(dummies[0]) == ncols:
                break
        x_new = np.dot(dummies, scaling)
        return x_new, np.var(x_new)

    def scale(self, lv: str, mv: str, z_by_lv: np.ndarray, weights) -> pd.DataFrame:
        z_by_lv = weights.get_Z_for_mode_b(lv, mv, z_by_lv)
        to_quantify = np.array([weights.mv_grouped_by_lv(lv, mv), z_by_lv])
        quantified = util.groupby_mean(to_quantify)
        x_quant_incr, var_incr = self._ordinalize(quantified[1, :], weights.dummies(mv).copy(), z_by_lv, 1)
        x_quant_decr, var_decr = self._ordinalize(quantified[1, :], weights.dummies(mv).copy(), z_by_lv, -1)
        x_quantified = -x_quant_decr if var_incr < var_decr else x_quant_incr
        scaled = util.treat_numpy(x_quantified) * weights.correction()
        return scaled


class _Nominal:

    def scale(self, lv: str, mv: str, z_by_lv: np.ndarray, weights) -> pd.DataFrame:
        z_by_lv = weights.get_Z_for_mode_b(lv, mv, z_by_lv)
        to_quantify = np.array([weights.mv_grouped_by_lv(lv, mv), z_by_lv])
        quantified = util.groupby_mean(to_quantify)
        x_quantified = weights.dummies(mv).dot(quantified[1, :])
        return util.treat_numpy(x_quantified) * weights.correction()


class Scale(Enum):
    """
    Used to specify the measurement type of a manifest variable when performing calculations with nonmetric data.

    * :attr:`RAW` for numerical variables that require no transformation;
    * :attr:`NUM` for numerical variables that are suitable for linear transformation;
    * :attr:`ORD` for ordinal variables that are suitable for monotonic transformation;
    * :attr:`NOM` for nominal variables that are suitable for non-monotonic transformation.
    """
    NUM = _Numeric()
    RAW = _Raw()
    ORD = _Ordinal()
    NOM = _Nominal()



#!/usr/bin/python3
#
# Copyright (C) 2019 Google Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import statsmodels.api as sm, numpy as np, pandas as pd
from enum import Enum


class _CentroidInnerWeightCalculator:

    def calculate(self, path: pd.DataFrame, y: np.ndarray) -> np.ndarray:
        return np.sign(np.corrcoef(y, rowvar=False) * (path + path.transpose()))


class _FactorialInnerWeightCalculator:

    def calculate(self, path: pd.DataFrame, y: np.ndarray) -> np.ndarray:
        return np.cov(y, rowvar=False) * (path + path.transpose())


class _PathInnerWeightCalculator:

    def calculate(self, path: pd.DataFrame, y: np.ndarray) -> np.ndarray:
        E = path.values.astype(np.float64)
        for i in range(E.shape[0]):
            follow = path.iloc[i, :] == 1
            if path.iloc[i, :].sum() > 0:
                E[follow, i] = sm.OLS(y[:, i], y[:, follow]).fit().params
            predec = path.iloc[:, i] == 1
            if path.iloc[:, i].sum() > 0:
                E[predec, i] = np.corrcoef(np.column_stack((y[:, predec], y[:, i])), rowvar=False)[:,-1][:-1]
        return E


class Scheme(Enum):
    """
    The scheme to use to calculate inner weights.
    """
    CENTROID = _CentroidInnerWeightCalculator()
    PATH = _PathInnerWeightCalculator()
    FACTORIAL = _FactorialInnerWeightCalculator()



#!/usr/bin/python3
#
# Copyright (C) 2019 Google Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pandas as pd, numpy as np, plspm.util as util
from sklearn.decomposition import PCA
from plspm.config import Config
from plspm.mode import Mode

class Unidimensionality:
    """Internal class that computes various reliability metrics. Use the method :meth:`~.plspm.Plspm.unidimensionality` defined on :class:`~.plspm.Plspm` to retrieve the results."""
    def __init__(self, config: Config, data: pd.DataFrame, correction: float):
        self.__config = config
        self.__data = data
        self.__correction = correction

    def summary(self):
        """Internal method that performs principal component analysis to compute various reliability metrics."""
        summary = pd.DataFrame(np.NaN, index=self.__config.lvs(),
                               columns=["mode", "mvs", "cronbach_alpha", "dillon_goldstein_rho", "eig_1st", "eig_2nd"])
        for lv in self.__config.lvs():
            mvs = len(self.__config.mvs(lv))
            summary.loc[lv, "mode"] = self.__config.mode(lv).name
            summary.loc[lv, "mvs"] = mvs
            if not self.__data.loc[:,self.__config.mvs(lv)].isnull().values.any():
                mvs_for_lvs = util.treat(self.__data.filter(self.__config.mvs(lv))) * self.__correction
                pca_input = mvs_for_lvs if mvs_for_lvs.shape[0] > mvs_for_lvs.shape[1] else mvs_for_lvs.transpose()
                pca = PCA()
                pca_scores = pca.fit_transform(pca_input)
                pca_std_dev = np.std(pca_scores, axis=0)
                summary.loc[lv, "eig_1st"] = pca_std_dev[0] ** 2
                summary.loc[lv, "eig_2nd"] = pca_std_dev[1] ** 2 if mvs > 1 else np.nan
                if (self.__config.mode(lv) == Mode.A):
                    if mvs > 1:
                        ca_numerator = 2 * np.tril(pca_input.corr(), -1).sum()
                        ca_denominator = pca_input.sum(axis=1).var() / self.__correction ** 2
                        ca = max(0, (ca_numerator / ca_denominator) * (mvs / (mvs - 1)))
                    else:
                        ca = np.nan
                    summary.loc[lv, "cronbach_alpha"] = ca
                    corr = np.corrcoef(np.column_stack((pca_input.values, pca_scores[:,0])), rowvar=False)[:,-1][:-1]
                    rho_numerator = sum(corr) ** 2
                    rho_denominator = rho_numerator + (mvs - np.sum(np.power(corr, 2)))
                    summary.loc[lv, "dillon_goldstein_rho"] = rho_numerator / rho_denominator
        return summary



#!/usr/bin/python3
#
# Copyright (C) 2019 Google Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pandas as pd, numpy as np
from plspm.config import Config
from plspm.mode import Mode


class InnerSummary:
    """Internal class that computes a summary of the inner model.  Use the methods :meth:`~plspm.Plspm.inner_summary` and :meth:`~plspm.Plspm.goodness_of_fit` defined on :class:`~.plspm.Plspm` to retrieve the inner model characteristics."""

    def __init__(self, config: Config, r_squared: pd.Series, r_squared_adj: pd.Series, outer_model: pd.DataFrame):
        path = config.path()
        lv_type = path.sum(axis=1).astype(bool)
        lv_type.name = "type"
        lv_type_text = lv_type.replace(False, "Exogenous").replace(True, "Endogenous")
        block_communality = pd.Series(0, index=path.index, name="block_communality")
        mean_redundancy = pd.Series(0, index=path.index, name="mean_redundancy")
        ave = pd.Series(0, index=path.index, name="ave")
        communality_aux = []
        num_mvs_in_lv = []
        for lv in config.lvs():
            communality = outer_model.loc[:, "communality"].loc[config.mvs(lv)]
            block_communality.loc[lv] = communality.mean()
            mean_redundancy.loc[lv] = outer_model.loc[:, "redundancy"].loc[config.mvs(lv)].mean()
            if config.mode(lv) == Mode.A:
                ave_numerator = communality.sum()
                ave_denominator = ave_numerator + (1 - communality).sum()
                ave.loc[lv] = ave_numerator / ave_denominator
            if len(config.mvs(lv)) > 1:
                num_mvs_in_lv.append(len(config.mvs(lv)))
                communality_aux.append(block_communality.loc[lv])
        self.__summary = pd.concat([lv_type_text, r_squared, r_squared_adj, block_communality, mean_redundancy, ave], axis=1,
                                   sort=True)
        mean_communality = sum(x * y for x, y in zip(communality_aux, num_mvs_in_lv)) / sum(num_mvs_in_lv)
        r_squared_aux = r_squared * lv_type
        self.__goodness_of_fit = np.sqrt(mean_communality * r_squared_aux[r_squared_aux != 0].mean())

    def summary(self) -> pd.DataFrame:
        """Internal method that returns the summary of the inner model."""
        return self.__summary

    def goodness_of_fit(self) -> float:
        """Internal method that returns the goodness-of-fit of the model."""
        return self.__goodness_of_fit
